import pandas as pd
import numpy as np


def flatten(rho, pval, sortby="cor"):
    """Flatten correlation and p-value matrix

    Parameters:
    -----------
    rho : ndarray
        Correlation Matrix

    pval : ndarray
        Matrix with p-values

    sortby : str
        sort the output table by
        - "cor"  the highest absolute correlation coefficient
        - "pval" the lowest p-value

    Return:
    -------
    tab : ndarray
        Table with (i, j, cor, pval) rows

    Example:
    --------
        from korr import pearson, flatten
        rho, pval = pearson(X)
        tab = flatten(rho, pval, sortby="pval")
        tab.values
    """
    n = rho.shape[0]

    idx = np.triu_indices(n, k=1)

    tab = pd.DataFrame(
        columns=['i', 'j', 'cor', 'pval'],
        data=np.c_[idx[0], idx[1], rho[idx], pval[idx]])
    tab[['i', "j"]] = tab[['i', "j"]].astype(int)

    if sortby == "cor":
        tab['abscor'] = np.abs(tab['cor'])
        tab.sort_values(by='abscor', inplace=True, ascending=False)
    elif sortby == "pval":
        tab.sort_values(by='pval', inplace=True, ascending=True)

    return tab[["i", "j", "cor", "pval"]]



__version__ = '0.8.3'

from .pearson import pearson
from .kendall import kendall
from .spearman import spearman
from .find_unrelated import find_unrelated
from .flatten import flatten
from .corrgram import corrgram
from .corr_vs_pval import corr_vs_pval
from .bracket_pval import bracket_pval
from .confusion import confusion
from .mcc import mcc, confusion_to_mcc
from .slice_yx import slice_yx
from .find_best import find_best
from .find_worst import find_worst
from .mincorr import mincorr, mincorr_result_check
from .bootcorr import bootcorr




import matplotlib.pyplot as plt
import matplotlib
import numpy as np


def corrgram(r, pval, varnames=None, dpi=96, cmap='RdBu'):
    # cmap: RdBu, PuOr, coolwarm

    # use numbers as
    if varnames is None:
        varnames = [str(i) for i in range(r.shape[1])]

    # number of variables
    num = len(varnames)

    # create figure and axes objects
    fig, ax = plt.subplots(dpi=dpi, figsize=(num * .75, num * .75))
    # fig.tight_layout()

    # plot heatmap
    im = ax.imshow(r, cmap=cmap, interpolation='nearest', vmin=-1, vmax=1)

    # axis formating
    plt.yticks(range(num), varnames)
    plt.xticks(range(num), varnames, rotation=45,
               ha="left", rotation_mode="anchor")
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)
    # annotate axes
    ax.set_xticks(np.arange(num + 1) - .5, minor=True)
    ax.set_yticks(np.arange(num + 1) - .5, minor=True)
    ax.tick_params(which="minor", bottom=False, left=False)

    # Turn spines off and create white grid.
    ax.grid(which="minor", color="white", linestyle='-', linewidth=3)
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    # colorbar
    cbar = ax.figure.colorbar(im, fraction=0.046, pad=0.04)
    matplotlib.cm.ScalarMappable().set_clim(vmin=-1, vmax=+1)
    cbar.ax.set_ylabel('correlation coefficient', rotation=-90, va="bottom")
    cbar.outline.set_edgecolor('darkgray')

    # add text
    for i in range(num):
        for j in range(num):
            if i is not j:
                # concate the text
                strtxt = (
                    '+' if r[i, j] > 0 else ('-' if r[i, j] < 0 else '')) \
                    + "." + '{:02d}'.format(int(abs(r[i, j]) * 100)) \
                    + '\n(.{:03d})'.format(int(pval[i, j] * 1000))
                # color depending on the absolute corr coeff value
                strcol = "gray" if abs(r[i, j]) < 0.4 else "white"
                # add the text to the graph
                ax.text(j, i, strtxt, ha="center",
                        va="center", color=strcol)

    # white bg
    fig.patch.set_facecolor('white')

    # for later usage
    return fig, ax, cbar



import matplotlib.pyplot as plt
import numpy as np


def bracket_pval(r, pval, plim=[.1, .01, .001], nbins=None, dpi=96):
    # reshape if r and pval are matricies
    # otherwise corr[i,j] would be counted twice
    if len(r.shape) == 2:
        idx = (np.tri(N=r.shape[0], k=-1) == 1)
        r = r[idx]
        pval = pval[idx]

    # Use nbins=20 from [-1,+1] for correlations
    if nbins is None:
        absmax = np.max(np.abs(r))
        nbins = (np.arange(0, 21) / 10 - 1) * absmax

    # set limits for p-values
    if isinstance(plim, list):
        plim = sorted(plim, reverse=True) + [0]
    elif isinstance(plim, int):
        plim = [pow(10, -i) for i in range(1, plim + 1)] + [0]

    # create subsets
    xlist = []
    names = []
    # add the subset pval=1 to the first plim item
    xlist.append(r[(pval >= plim[0])])
    names.append('p >= ' + str(plim[0]))
    # add the rest
    for i in range(1, len(plim)):
        pup = plim[i - 1]
        plo = plim[i]
        xlist.append(r[(pup > pval) & (pval >= plo)])
        names.append(str(pup) + ' > p >= ' + str(plo))

    # create histogram
    fig, ax = plt.subplots(dpi=dpi)
    ax.hist(xlist, histtype='bar', stacked=True, bins=nbins)

    # annotation
    ax.legend(names, loc=2, bbox_to_anchor=(1.04, 1.04))
    ax.set_ylabel('frequency')
    ax.set_xlabel('correlation coefficient')

    # design grid
    ax.grid(color='darkgray', linestyle='-.')
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    # output figure objects
    return fig, ax, xlist, names



import warnings
import numpy as np


def find_best(rho, pval, m=1, rlim=.4, plim=.01):
    """Find the N "best", i.e. high and most significant, correlations

    Parameters
    ----------
    rho : ndarray, list
        1D array with correlation coefficients

    pval : ndarray, list
        1D array with p-values

    m : int
        The desired number of indicies to return
        (How many "best" correlations to find?)

    rlim : float
        Desired minimum absolute correlation coefficient
        (Default: 0.40)

    plim : float
        Desired maximum p-value
        (Default: 0.01)

    Return
    ------
    selected : list
        Indicies of rho and pval of the "best" correlations.
    """
    # convert to lists
    n = len(rho)
    r = list(np.abs(rho))
    p = list(pval)
    i = list(range(n))

    # check m
    if m > n:
        warnings.warn(
            'm is bigger than the available correlations in rho and pval.')
        m = n

    # selected indicies
    selected = list()
    # (1) pick the highest |r| that are
    #   |r| >= r_lim
    #    p  <  p_lim
    it = 0
    while (len(selected) < m) and (it < n):
        # find the best |r| of the remaining values
        temp = r.index(max(r))  # temporary index of the remaining values
        best = i[temp]  # store original index  as 'best' before abort loop
        # check
        if (r[temp] >= rlim) and (p[temp] < plim):
            # delete from lists
            r.pop(temp)
            p.pop(temp)
            i.pop(temp)
            # append to abort
            selected.append(best)
        # next step
        it = it + 1
    # print(selected, i)

    # (2) Just pick the highest |r| of the remaining
    # with substantial p-values
    #    p  <  p_lim
    it = 0
    n2 = len(i)
    while (len(selected) < m) and (it < n2):
        # find the best |r| of the remaining values
        temp = r.index(max(r))  # temporary index of the remaining values
        best = i[temp]  # store original index  as 'best' before abort loop
        # check
        if (p[temp] < plim):
            # delete from lists
            r.pop(temp)
            p.pop(temp)
            i.pop(temp)
            # append to abort
            selected.append(best)
        # next step
        it = it + 1

    # (3) Pick the lowest p-values
    # If the correlations aren't useful, then prefer the p-values
    it = 0
    n3 = len(i)
    while (len(selected) < m) and (it < n3):
        # find the smallest p-value
        temp = p.index(min(p))
        best = i[temp]
        # delete from lists
        r.pop(temp)
        p.pop(temp)
        i.pop(temp)
        # append to abort
        selected.append(best)
        # next step
        it = it + 1

    return selected



import numpy as np


def confusion(a, p):
    """Confusion Matrix

    - confusion matrix, error matrix, matching matrix (unsupervised)
    - is a special case: 2x2 contingency table (xtab, RxC table)
    - both dimensions are variables with the same classes/labels,
      i.e. actual and predicted variables are binary [0,1]

    Parameters:
    -----------
    a : ndarray
        Actual values, binary [0,1]

    p : ndarray
        Predicted values, binary [0,1]

    Returns:
    --------
    cm : ndarray
        Confusion matrix

                       predicted=0   predicted=1
            actual=0        tn           fp
            actual=1        fn           tp

    Example:
    --------
        import korr
        cm = korr.confusion(a, p)
        tn, fp, fn, tp = cm.ravel()

    Alternatives:
    -------------
        import pandas as pd
        cm = pd.crosstab(a, p)

        from sklearn.metrics import confusion_matrix
        cm = confusion_matrix(a, p)
    """
    m = a == p  # matches (a=1, p=1) and (a=0, p=0)
    f = np.logical_not(m)  # xor (a=1, p=0) and (a=0, p=1)
    tp = np.sum(np.logical_and(m, a))  # 11
    tn = np.sum(np.logical_and(m, np.logical_not(a)))  # 00
    fn = np.sum(np.logical_and(f, a))  # 10
    fp = np.sum(np.logical_and(f, p))  # 01
    return np.array([[tn, fp], [fn, tp]])



import numpy as np


def mincorr_cost_func(tmp, cost_func):
    if cost_func in ['sum', 'avg']:
        # corr sum of each "available" k to the "selected" i
        sk = np.sum(tmp, axis=0)
    elif cost_func in ['sqrt', 'squared']:
        sk = np.sum(np.power(tmp, 2), axis=0)
    elif cost_func in ['mu_max']:
        sk = np.mean(tmp, axis=0) + np.max(tmp, axis=0)
    return sk


def mincorr(cmat, n_stop=5, max_rho=0.4, init_method='avg',
            cost_fn='sqrt', verbose=True):
    # declare list of indicies
    available = list(range(cmat.shape[0]))
    selected = []

    # prep: convert to abs correlations
    absrho = np.abs(cmat, order='C')

    # (1) Initial Condition
    # (1a) find "i" with the lowest correlations
    if init_method in ['sum', 'avg']:
        si = np.sum(absrho, axis=1)
    elif init_method in ['low', 'lowest', 'min']:
        si = np.min(absrho, axis=1)  # same as picking the lowest |r_ij|
    else:
        raise Exception(
            "unknown init_method='{:s}' value".format(str(init_method)))

    i = np.argmin(si)
    selected.append(i)
    available.remove(i)

    # (1b) find "j" with the lowest correlation |r_ij| given "i"
    j = np.argmin(absrho[i, :])
    selected.append(j)
    available.remove(j)

    # (2) Enforce max abs correlation
    if max_rho < 1.0:
        # (2a) reset all absrho>maxrho to np.inf
        absrho[absrho > max_rho] = np.inf

        # (2b) add features if corr to previously
        # selected features is below max_rho
        for _ in range(2, n_stop):
            # slice available
            tmp = absrho[selected, :]  # slice \r_ij| given i \in "selected"
            tmp = tmp[:, available]  # remove "selected" colums (with 1s)
            # find best
            # sk = np.sum(np.power(tmp,2), axis=0)
            sk = mincorr_cost_func(tmp, cost_fn)
            k = np.argmin(sk)   # find index with lowest cost func
            if sk[k] < np.inf:
                # store indicies
                j = available[k]  # get the index
                selected.append(j)
                available.remove(j)
            else:
                if verbose:
                    print(("Terminated: max_rho={:5.3f} works only "
                           "for {:d} features".format(max_rho, _)))
                break
        # (2c) Reset matrix
        absrho = np.abs(cmat, order='C')

    # (3) fill the remaining spots
    for _ in range(len(selected), n_stop):
        # slice available
        tmp = absrho[selected, :]  # slice \r_ij| given i \in "selected"
        tmp = tmp[:, available]  # remove "selected" colums (with 1s)
        # find best
        # sk = np.sum(np.power(tmp,2), axis=0)
        sk = mincorr_cost_func(tmp, cost_fn)
        k = np.argmin(sk)  # find index with lowest cost func
        # store indicies
        j = available[k]  # get the index
        selected.append(j)
        available.remove(j)

    # done
    return selected


def mincorr_result_check(cmat, selected):
    print("selected: ", selected)
    print("avg abs corr: {:7.4f}".format(
        np.sum(np.abs(np.tril(cmat[selected, :][:, selected], -1))) / (
            (len(selected) - 1) * len(selected) / 2)))
    print("selected corr matrix:\n", cmat[selected, :][:, selected].round(3))
    print("")



import scipy.stats  # pearsonr
import numpy as np


def find_unrelated(x, plim=0.1, axis=0):
    """Find indicies of insignificant un-/correlated variables

    Example:
    --------
        i, j = find_unrelated(x, plim, rlim)

    """
    # transpose if axis<>0
    if axis != 0:
        x = x.T

    # read dimensions and allocate variables
    _, c = x.shape
    pairs = []

    # compute each (i,j)-th correlation
    for i in range(0, c):
        for j in range(i + 1, c):
            _, p = scipy.stats.pearsonr(x[:, i], x[:, j])
            if p > plim:
                pairs.append((i, j))
    # done
    return tuple(pairs)




def slice_yx(r, pval, ydim=1):
    """slice a correlation and p-value matrix of a (y,X) dataset
    into a (y,x_i) vector and (x_j, x_k) matrices

    Parameters
    ----------
    r : ndarray
        Correlation Matrix of a (y,X) dataset

    pval : ndarray
        p-values

    ydim : int
        Number of target variables y, i.e. the first ydim-th columns
        and rows are (y, x_i) correlations

    Returns
    -------
    y_r : ndarray
        1D vector or ydim-column array with (y,x) correlations

    y_pval : ndarray
        1D vector or ydim-column array with (y,x) p-values

    x_r : ndarray
        correlation matrix (x_j, x_k)

    x_pval : ndarar
        matrix with p-values

    Example
    -------
        import korr
        r, pval = korr.mcc(np.c_[y, X])
        y_r, y_pval, x_r, x_pval = slice_yx(r, pval, ydim=1)
        print(np.c_[y_r, y_pval])
        korr.corrgram(x_r, x_pval)
    """
    if ydim == 1:
        return (
            r[1:, :1].reshape(-1, ), pval[1:, :1].reshape(-1, ),
            r[1:, 1:], pval[1:, 1:])
    else:
        return (
            r[ydim:, :ydim], pval[ydim:, :ydim],
            r[ydim:, ydim:], pval[ydim:, ydim:])



import matplotlib.pyplot as plt
import numpy as np


def corr_vs_pval(r, pval, plim=0.01, rlim=0.4, dpi=96):
    """Histogram for correlation coefficients and its p-values (colored)

    Parameters:
    -----------
    r : np.ndarray
        Correlation coefficient matrix.
        The upper triangular elements are extracted if a NxN is provided.
        Otherwise provide a plain vector.

    pval : np.ndarray
        NxN matrix with p-values

    plim : float
        Discretionary alpha threshold to judge if a p-value is considered
        "significant" or not. (Default: 0.01 or 1%)

    rlim : float
        Descretionary threshold to judge if an absolute correlation
        coefficient is big enough. (Default: 0.4)

    dpi : int
        Set the resolution of the matplotlib graphics.

    Return:
    -------
    fig, ax
        matplotlib figure and axes for further tweaking
    """
    # reshape
    if len(r.shape) == 2:
        idx = (np.tri(N=r.shape[0], k=-1) == 1)
        r = r[idx]
        pval = pval[idx]

    # indicies for the three groups
    i1 = (pval >= plim)
    i2 = (pval < plim) & (np.abs(r) > rlim)
    i3 = (pval < plim) & (np.abs(r) <= rlim)

    # plot paramters
    absmax = np.max(np.abs(r))
    b = (np.arange(0, 21) / 10 - 1) * absmax
    c = plt.get_cmap('tab10').colors
    c = (c[1], c[8], c[7])

    # create plot
    fig, ax = plt.subplots(dpi=dpi)
    ax.hist((r[i1], r[i2], r[i3]), histtype='bar',
            stacked=True, bins=b, color=c)

    # legend, title, labels
    ax.legend(['p >= ' + str(plim),
               'p < ' + str(plim) + ' and |r| > ' + str(rlim),
               'p < ' + str(plim) + ' and |r| <= ' + str(rlim)],
              loc=2, bbox_to_anchor=(1.04, 1.04))
    ax.set_ylabel('frequency')
    ax.set_xlabel('correlation coefficient')

    # design grid
    ax.grid(color='darkgray', linestyle='-.')
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    # output figure/axes onbjects
    return fig, ax



from setuptools import setup
import pypandoc


def get_version(path):
    with open(path, "r") as fp:
        lines = fp.read()
    for line in lines.split("\n"):
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(name='korr',
      version=get_version("korr/__init__.py"),
      description='collection of utility functions for correlation analysis',
      long_description=pypandoc.convert('README.md', 'rst'),
      url='http://github.com/kmedian/korr',
      author='Ulf Hamster',
      author_email='554c46@gmail.com',
      license='MIT',
      packages=['korr'],
      install_requires=[
          'setuptools>=40.0.0',
          'numpy>=1.14.5,<2',
          'scipy>=1.1.0,<2',
          'matplotlib>=3.3.0'],
      python_requires='>=3.6',
      zip_safe=True)



import scipy.stats  # kendalltau
import numpy as np


def kendall(x, axis=0):
    """Kendall' tau (Rank) Correlation Matrix (for ordinal data)

    Parameters
    ----------
    x : ndarray
        data set

    axis : int, optional
        Variables as columns is the default (axis=0). If variables are
        in the rows use axis=1

    Returns
    -------
    r : ndarray
        Correlation Matrix (Kendall tau)

    p : ndarray
        p-values
    """
    # transpose if axis<>0
    if axis != 0:
        x = x.T

    # read dimensions and
    n, c = x.shape

    # check if enough variables provided
    if c < 2:
        raise Exception(
            "Only " + str(c) + " variables provided. Min. 2 required.")

    # allocate variables
    r = np.ones((c, c))
    p = np.zeros((c, c))

    # compute each (i,j)-th correlation
    for i in range(0, c):
        for j in range(i + 1, c):
            r[i, j], p[i, j] = scipy.stats.kendalltau(x[:, i], x[:, j])
            r[j, i] = r[i, j]
            p[j, i] = p[i, j]

    # done
    return r, p



from .confusion import confusion
import scipy.stats
import numpy as np


def confusion_to_mcc(*args):
    """Convert the confusion matrix to the Matthews correlation coefficient

    Parameters:
    -----------
    cm : ndarray
        2x2 confusion matrix with np.array([[tn, fp], [fn, tp]])

    tn, fp, fn, tp : float
        four scalar variables
        - tn : number of true negatives
        - fp : number of false positives
        - fn : number of false negatives
        - tp : number of true positives

    Return:
    -------
    r : float
        Matthews correlation coefficient
    """
    if len(args) == 1:
        tn, fp, fn, tp = args[0].ravel().astype(float)
    elif len(args) == 4:
        tn, fp, fn, tp = [float(a) for a in args]
    else:
        raise Exception((
            "Input argument is not an 2x2 matrix, "
            "nor 4 elements tn, fp, fn, tp."))

    return (tp * tn - fp * fn) / np.sqrt(
        (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))


def mcc(x, axis=0, autocorrect=False):
    """Matthews correlation

    Parameters
    ----------
    x : ndarray
        dataset of binary [0,1] values

    axis : int, optional
        Variables as columns is the default (axis=0). If variables
        are in the rows use axis=1

    autocorrect : bool, optional
        If all predictions are True or all are False, then MCC
        returns np.NaN
        Set autocorrect=True to return a 0.0 correlation instead.

    Returns
    -------
    r : ndarray
        Matthews correlation

    p : ndarray
        p-values of the Chi^2 test statistics

    Notes:
    ------
    (1) We cannot directly transform the Chi^2 test statistics to
    the Matthews correlation because the relationship is

         |r| = sqrt(chi2 / n)
         chi2 = r * r * n

    (2) The sign would be missing. Therefore, as a rule of thumbs,
    If you want to optimize ABS(r_mcc) then just use the Chi2/n
    directly (Divide Chi^2 by the number of observations)

    Examples:
    ---------
        import korr
        r, pval = korr.mcc(X)

    Alternatives:
    -------------
        from sklearn.metrics import matthews_corrcoef
        r = matthews_corrcoef(y_true, y_pred)
    """
    # transpose if axis<>0
    if axis != 0:
        x = x.T

    # read dimensions and
    n, c = x.shape

    # check if enough variables provided
    if c < 2:
        raise Exception(
            "Only " + str(c) + " variables provided. Min. 2 required.")

    # allocate variables
    r = np.ones((c, c))
    p = np.zeros((c, c))

    # compute each (i,j)-th correlation
    for i in range(0, c):
        for j in range(i + 1, c):
            cm = confusion(x[:, i], x[:, j])
            r[i, j] = confusion_to_mcc(cm)
            r[j, i] = r[i, j]
            p[i, j] = 1 - scipy.stats.chi2.cdf(r[i, j] * r[i, j] * n, 1)
            p[j, i] = p[i, j]

    # replace NaN with 0.0
    if autocorrect:
        r = np.nan_to_num(r)

    # done
    return r, p



import numpy as np
import korr


def bootcorr(X, n_draws=30, subsample=0.7, replace=True,
             random_state=42, corr_fn=korr.pearson):
    """Estimate multiple correlation matrices based on
    bootstrapped samples.

    The produced correlation matrices can be used to
    check if certain correlation estimates are stable.

    Parameters
    ----------
    n_draws : int
        number of bootstraps.

    subsample : float
        percentage of samples to use in every estimation.

    replace: bool
        True (default) with replacement, i.e. samples can be used multiple
        times (bootstrapping).
        False is without replacement, i.e. samples are used once, sample are
        not reused (SRSWOR)

    corr_fn : function
        python function that returns a correlation matrix and matrix with
        corresponding p-values
    """
    # number of examples
    n_samples, n_features = X.shape

    # size
    if isinstance(subsample, int):
        n_size = subsample
    else:
        n_size = int(n_samples * subsample)

    # create empty tensors
    rho3 = np.zeros(shape=(n_draws, n_features, n_features))
    rho3[:] = np.nan

    pval3 = np.zeros(shape=(n_draws, n_features, n_features))
    pval3[:] = np.nan

    oob = []
    rng = set(range(len(X)))

    # set seed
    np.random.seed(random_state)

    for d in range(n_draws):
        idx = np.random.choice(range(n_samples), size=n_size, replace=replace)
        oob.append(list(rng - set(idx)))
        rho3[d], pval3[d] = corr_fn(X[idx, :])

    # done
    return rho3, pval3, oob
